/* Tool for working with BIF-6809 images.
// Written by Joel Matthew Rees, Amagasaki, Japan, April 2019,
// Parts adapted from the author's 32col.c, written 1999.
// Copyright 1999, 2019, Joel Matthew Rees.
// Permission granted in advance for all uses
// with the condition that this copyright and permission notice are retained.
//
// BIF-6809 project page: https://osdn.net/projects/bif-6809/
*/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>  /* for EXIT_SUCCESS */
#include <string.h>
#include <ctype.h>


#define kScreenSize 1024
#define kScreenWidth 32
#define kScreenHeight ( kScreenSize / kScreenWidth )
#define kBufferPlay 3  /* room for CR/LF and NUL */
#define kBufferWidth ( kScreenWidth + kBufferPlay ) /* Should never be used. */


#define TO_SCREEN 1
const char kTo_ScreenStr[] = "--to-screens";
#define TO_EOLN_TEXT 2
const char kTo_EOLN_textStr[] = "--to-eoln-text";

const char kBlockSizeStr[] = "-size";
const char kBlockWidthStr[] = "-width";
const char kBlockOffsetStr[] = "-off";
const char kBlockCountStr[] = "-count";
const char kSuppressEndLinesStr[] = "-suppressEndLines";

void toEOLNtext( FILE * input, FILE * output, char * buffer /* Must have room for kBufferPlay extra bytes per line. */,
                 unsigned blocksize, unsigned width, unsigned offset, unsigned count,
                 int suppressEndLines /*, int linecountflag */ )
{
  unsigned long start = blocksize * offset;
  unsigned long bytecount = blocksize * count;
  unsigned long totalBytes = 0;
  unsigned screenHeight = blocksize / width;
  unsigned bufferWidth = width + kBufferPlay;

/* dbg */ fprintf( stderr, "size: %u; width: %u; off: %u; count: %u\n", blocksize, width, offset, count ); 

  if ( start > 0 )
  { fseek( input, start, SEEK_SET );
  }
  while ( !feof( input ) && ( totalBytes < bytecount ) )
  {
    int lineCount;
    for ( lineCount = 0; lineCount < screenHeight && !feof( input ); ++lineCount )
    {
      char * linestart = buffer + lineCount * bufferWidth;
      int length = fread( linestart, sizeof (char), width, input );
      totalBytes += length;
      while ( --length >= 0 
          && ( isspace( linestart[ length ] ) 
             || !isprint( linestart[ length ] ) ) )
        /* "empty" loop */;
      linestart[ ++length ] = '\0';
    }
    if ( lineCount > 1 
       || ( lineCount == 1 && buffer[ 0 ] != '\0' ) )
    {
      int line = 0;
      if ( suppressEndLines )
      { while ( --lineCount > 0 && buffer[ lineCount * bufferWidth ] == '\0' ) 
          { /* "empty" loop: note tested NUL is first character of line. */ }
      }
      else 
      { --lineCount;
      }
      for ( line = 0; line <= lineCount; ++line ) /* End condition intentional! */
      { fputs( buffer + line * bufferWidth, output );
        fputc( '\n', output );
      }
      /* fputc( '\f', output ); This is not useful. */
    }
  }
}


#define OVER_CHAR 0x100		/* char range boundary + 1 */
#define FILE_START 0x200	/* beyond char range. */
#define LINE_START 0x400	/* beyond char range. */
#define LINE_START32 0x800	/* beyond char range. */


void toScreens( FILE * input, FILE * output, char * buffer /* Must have room for kBufferPlay extra bytes per line. */,
                 unsigned blocksize, unsigned width, unsigned offset, unsigned count
                 /*, int linecountflag */ )
{
  unsigned long start = blocksize * offset;
  unsigned screenHeight = ( blocksize / width );
  unsigned bufferWidth = width + kBufferPlay;
  int eolFlag = FILE_START;

  if ( start > 0 )
  { fseek( output, start, SEEK_SET );
  }
  while ( !feof( input ) )
  {
    int lineCount;
    for ( lineCount = 0; lineCount < screenHeight; ++lineCount )
    {
      int length = 0;
      char * line = buffer + lineCount * bufferWidth;
      int ch = LINE_START;
      while ( ( length < width ) && !feof( input ) )
      { ch = fgetc( input );
        if ( length == 0 )
        { if ( ( eolFlag < OVER_CHAR ) /* EOL did not come before 32nd. */
               && ( ( ch == '\n' ) || ( ch == '\r' ) ) )
          { eolFlag = ch; /* But throw ch away and get another. */
            ch = fgetc( input );
          }
          if ( ( ( ch == '\r' ) && ( eolFlag == '\n' ) )
               || ( ( ch == '\n' ) && ( eolFlag == '\r' ) ) )
          { /* Throw ch away and get another; eolFlag has done its job. */
            ch = fgetc( input );
          }
        }
        eolFlag = ch; /* At this point, ch is validly the first character of the line. */
        if ( ( ch == '\n' ) || ( ch == '\r' ) || feof( input ) )
        { eolFlag = LINE_START;
          break;	/* The habit is to set a NUL, but not for SCREENs. */
        }
        line[ length++ ] = ch;
/* dbg * / fputc( ch, stderr ); */
      }
/* dbg * / fprintf( stderr, "||end:%d:", length ); */
      while ( length < width )
      { line[ length++ ] = ' ';
/* dbg * / fputc( '*', stderr ); */
      }
/* dbg * / fprintf( stderr, "||:%d:%d\n", length, lineCount ); */
    }
/* dbg * / fprintf( stderr, "<<screen:%d:>>\n", lineCount ); */
    if ( lineCount > 0 )
    { int line = 0;
      size_t count = 0;
      int error = 0;
      for ( line = 0; line < lineCount; ++line )
      { count = fwrite( buffer + line * bufferWidth, sizeof (char), width, output ); 
        if ( ( count != width ) || ( ( error = ferror( output ) ) != 0 ) )
        { fprintf( stderr, "Output error=%d; count: %lu::", error, count );
        }
/* dbg * / fprintf( stderr, "Output error=%d; count: %lu::", ferror( output ), count ); */
/* dbg * / { int i; for ( i = 0; i < width; ++i ) fputc( buffer[ line * bufferWidth + i ], stderr ); } */
/* dbg * / fputc( '\n', stderr ); */
      }
    }
  }
}


int getNumericParameter( const char parameter[], char * argstr,
                         unsigned long * rval, long low, unsigned long high )
{
  char * scanpt = argstr;
  unsigned long result = 0;
  size_t eqpt = strlen( parameter );
  if ( strncmp( parameter, argstr, eqpt ) == 0 ) 
  {
    if ( argstr[ eqpt ] != '=' )
    { printf( "\t%s needs '=' in '%s', ", parameter, argstr );
      return INT_MIN | 16;
    }
    ++eqpt;
    scanpt += eqpt;
    result = strtoul( scanpt, &scanpt, 0 );
    if ( scanpt <= argstr + eqpt )
    { printf( "\tBad %s value specified in '%s'\n,", parameter, argstr );
      return INT_MIN | 32;
    }
    if ( ( result < low ) || ( result > high ) )
    { fprintf( stderr, "\t%s value %lu out of range in '%s', try %lu\n,", parameter, result, argstr, * rval );
      return INT_MIN | 64;
    }
    * rval = result;
    return 1;
  }
  return 0;
}


int main(int argc, char * argv[] )
{
  FILE * input = stdin;
  FILE * output = stdout;
  char * buffer = NULL;
  int direction = 0;
  int errval = 0;
  unsigned long blocksize = kScreenSize;
  unsigned long width = kScreenWidth;
  unsigned long offset = 0;
  unsigned long count = UINT_MAX;
  unsigned long suppressEndLines = 0;
  int i;

  for ( i = 4; i < argc; ++i )
  { int berr = 0;
    int werr = 0;
    int oerr = 0;
    int cerr = 0;
    int serr = 0;
    if ( ( ( berr |= getNumericParameter( kBlockSizeStr, argv[ i ], &blocksize, 1, 0x8000UL ) ) > 0 )
         || ( ( werr |= getNumericParameter( kBlockWidthStr, argv[ i ], &width, 1, 1024 ) ) > 0 )
         || ( ( oerr |= getNumericParameter( kBlockOffsetStr, argv[ i ], &offset, 0, USHRT_MAX ) ) > 0 )
         || ( ( cerr |= getNumericParameter( kBlockCountStr, argv[ i ], &count, 1, USHRT_MAX ) ) > 0 ) 
         || ( ( serr |= getNumericParameter( kSuppressEndLinesStr, argv[ i ], &suppressEndLines, 0, 1 ) ) > 0 ) 
       )
    { /* empty */ }
    else
    { printf( "\tUnrecognized %s\n", argv[ i ] ); /* This isn't firing for gobbledygook. */
    }
    errval |= berr | werr | oerr | cerr | serr;
  }
  if ( ( blocksize % width ) != 0 )
  {
    errval |= INT_MIN | 1024;
    printf( "Block size %lu is not even multiple of edit width %lu.\n", blocksize, width );
  }
  if ( ( errval >= 0 ) && ( argc > 3 ) )
  {
    if ( strcmp( argv[ 1 ], kTo_ScreenStr ) == 0 )
    {
      direction = TO_SCREEN;
    }
    else if ( strcmp( argv[ 1 ], kTo_EOLN_textStr ) == 0 )
    {
      direction = TO_EOLN_TEXT;
    }
    if ( direction != 0 )
    {
      if ( strcmp( argv[ 2 ], "--" ) != 0 )
      {
        input = fopen( argv[ 2 ], "rb" );
      }
      if ( input == NULL )
      {
        fprintf( stderr, "Error opening file <%s> for input.\n", argv[ 2 ] );
        direction |= INT_MIN | 4;
      }
      if ( strcmp( argv[ 3 ], "--" ) != 0 )
      {
        output = fopen( argv[ 3 ], "r+b" );
      }
      if ( output == NULL )
      {
        fprintf( stderr, "Error opening file <%s> for output.\n", argv[ 3 ] );
        fclose( input );
        direction |= INT_MIN | 8;
      }
      if ( ( buffer = malloc( blocksize + kBufferPlay * ( blocksize / width ) ) ) == NULL )
      { fprintf( stderr, "Buffer allocation failure\n" );
        direction |= INT_MIN | 16;
      }
    }
  }
  if ( direction < -1 )
  {
    fprintf( stderr, "*** %s quitting. ***\n", argv[ 0 ] );
    return EXIT_FAILURE;
  }
  else if ( direction == 0 )
  {
    puts( "usage:" );
    printf( "\t%s %s <infile> <outfile> [ %s=<block-size> ] [ %s=<width> ] [ %s=<offset> ] [ %s=<count> ]\n", 
            argv[ 0 ], kTo_ScreenStr, kBlockSizeStr, kBlockWidthStr, kBlockOffsetStr, kBlockCountStr );
    printf( "\t%s %s <infile> <outfile> [ %s=<block-size> ] [ %s=<width> ] [ %s=<offset> ] [ %s=<count> ] [ %s={0|1} ]\n", 
            argv[ 0 ], kTo_EOLN_textStr, kBlockSizeStr, kBlockWidthStr, kBlockOffsetStr, kBlockCountStr, kSuppressEndLinesStr );
    printf( "** Default block size is %d, compatible with Forth SCREENs.\n", kScreenSize );
    printf( "** Default width is %d, compatible with Color Computer 1 & 2 text display.\n", kScreenWidth );
    printf( "** Default count is length of input file.\n" );
    printf( "** %s=1 to suppress trailing blank lines in SCREEN, default 0.\n", kSuppressEndLinesStr );
    printf( "** 0xhexadecimal and 0octal permitted for size, etc.\n" );
    printf( "** Replace <file> with -- for stdfiles in pipes\n" );
    /* printf( "\t%s --to-image <filename> <imagename> <offset>\n", argv[ 0 ] ); */
    return EXIT_SUCCESS;
  }

  switch ( direction )
  {
  case TO_SCREEN:
    toScreens( input, output, buffer, blocksize, width, offset, count );
    break;
  case TO_EOLN_TEXT:
    toEOLNtext( input, output, buffer, blocksize, width, offset, count, suppressEndLines );
    break;
  }
  if ( buffer != NULL )
    free( buffer );
  if ( output != stdout )
    fclose( output );
  if ( input != stdin )
    fclose( input );

  return EXIT_SUCCESS;
}

